
from .system_header import *
import time
from typing import List,Tuple,Dict,Union,Any



class BetterThread(threading.Thread):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
        self._external_stop_ = threading.Event()
        self._external_stop_.clear()
        self._ack_stop_ = threading.Event()
        self._ack_stop_.clear()
        self._ack_start_ = threading.Event()
        self._ack_start_.clear()
        self._better_stop_reason_ = None
        target = kwargs['target']
        if callable(target) and hasattr(target,'__qualname__'):
            self._tgt_name = target.__qualname__
        elif callable(target) and hasattr(target,'__name__'):
            self._tgt_name = target.__name__
        else:
            self._tgt_name = "?"

    def __str__(self):
        state = ((int(self._ack_start_.is_set()) << 0)
                +(int(self._ack_stop_.is_set()) << 1)
                +(int(self._external_stop_.is_set() << 2)))
        if state == 0:
            state = "Pending"
        elif state == 1:
            state = "Running"
        elif state in [2,3]:
            state = "Finished"
        elif state in [4,5]:
            state = 'Stopping'
        elif state in [6,7]:
            state = 'Stopped'
        out = f"<BetterThread: {state}, name: {self._name}, target: {self._tgt_name}>"
        return out
    
    def __repr__(self):
        return str(self)

    def stop(self,reason:str=None):
        if not self._external_stop_.is_set():
            self._external_stop_.set()
        if reason is not None:
            if self._better_stop_reason_:
                self._better_stop_reason_ = "\n".join([self._better_stop_reason_,reason])
            else:
                self._better_stop_reason_ = reason

    def check_for_stop(self):
        return self._external_stop_.is_set()

    def _stopped(self):
        self._ack_stop_.set()
    
    def is_stopped(self):
        return self._ack_stop_.is_set()

    def wait(self,timeout:Union[int,float]=None):
        wait_at = time.time()
        if timeout is None:
            continue_loop = lambda : True
        else:
            continue_loop = lambda : time.time() - wait_at < timeout
        while (continue_loop()):
            if not self.is_alive():
                self._stopped()
                return True
        return False

    def join(self):
        super().join()
        if not self._ack_stop_.is_set():
            self._stopped()

    def start(self):
        self._ack_start_.set()
        super().start()

    def run(self):
        self._ack_start_.set()
        super().run()
        self._stopped()

    @property
    def reason(self):
        if self.is_alive():
            return None
        return self._better_stop_reason_




from .system_header import *
from . import (
    ConnectionRole as CR,
    BetterConnection as BC,
    BetterProcess as BP,
    BetterThread as BT
)
from .better_connection import zmq_poll
import os,time
from queue import deque

PING = b'__ping__'
PONG = b'__pong__'

class ClientRole(IntFlag):
    MAIN=0 # Main process/thread
    CHDT=1 # Main process : thread
    CHDP=2 # Child process


class ClientConnection(BC):
    def __init__(self,
            endpoint:Union[str,List[str]],
            _ctx:zmq.Context=None):
        if not isinstance(endpoint,list):
            endpoint = [endpoint]
        addr = list()
        port = list()
        for ep in endpoint:
            if '://' in ep:
                _,ep = ep.split("://")
            a,p = ep.split(":")
            addr.append(a)
            port.append(p)
        super(ClientConnection,self).__init__(
                CR.TCP_ | CR.SEND | CR.RECV,
                addr,
                port,
                _ctx)

class ClientNetworking(object):
    def __init__(self,
            endpoint:Union[str,List[str]],
            using="proc",
            _ctx:zmq.Context=None):
        self.endpoints = endpoint
        self.mode = int(using=="thread")
        self._ctx = _ctx
        self._task_callback = None
        self.listener = None
        self._pid = os.getpid()
        self.reset()

    def set_callback(self,fcn):
        if not callable(fcn):
            raise ValueError("callback must be callable")
        self._task_callback = fcn

    def reset(self):
        if self.listener is not None:
            self.stop()
            self.wait()
        self._ping = False
        self.state_map = dict()
        self.task_map = dict()
        self.reply_map = dict()
        self._mpipe, self._cpipe, self._mcond, self._ccond, self._lthread = [None,]*5
        self._fcn = ClientRole.MAIN

    def stop(self,reason:str=None):
        if self.listener is not None:
            self.listener.stop(reason)
        if self._lthread is not None:
            self._lthread.stop()
    def disconnect(self):
        self.stop()
    def close(self):
        self.stop()

    def start(self):
        if self.listener is not None:
            return
        if self.mode == 0:
            # multi-process
            self._mpipe,self._cpipe = mp.Pipe()
            self._mcond,self._ccond = mp.Condition(),mp.Condition()
        self.listener = [BP,BT][self.mode](target=self._interface)
        self.listener.start()
        if self.mode == 0:
            response = self._mpipe.recv()
            self.endpoint = response['endpoint']
            if isinstance(self.endpoint,list):
                for ep in self.endpoint:
                    self._add_new_sid(ep)
            else:
                self._add_new_sid(ep)
            self._lthread = BT(target=self._monitor_pipe)
            self._lthread.start()
    def connect(self):
        self.start()


    def _interface(self):
        self._fcn = ClientRole.CHDT if self.mode else ClientRole.CHDP
        connection = ClientConnection(self.endpoints,None if not self.mode else self._ctx)
        connection.connect()
        self.endpoint = connection.endpoint
        if not self.mode:
            self._cpipe.send({'endpoint':self.endpoint})

        for sid in self.endpoint:
            self._add_new_sid(sid)


        ####################
        ## DEBUG
        ####################
        debug_timeout = 30.0
        ticks = 0
        last = time.time()

        while not self.listener.check_for_stop():
            if not self.mode:
                if self._cpipe.poll(0.001):
                    incoming = self._cpipe.recv()
                    # print("INCOMING:",incoming)
                    if not isinstance(incoming,dict):
                        continue
                    if 'ping' in incoming:
                        self._ping = incoming['ping']
                    elif 'sid' in incoming:
                        sid = incoming['sid']
                        if 'reply' in incoming:
                            priority = incoming['priority']
                            msg = incoming['reply']
                        if 'send' in incoming:
                            msg = incoming['send']
                            priority = 0
                        self._reply(sid,msg,priority)                        

            if self._ping:
                for k,d_q in self.reply_map.items():
                    if d_q and d_q[0] == PING:
                        continue
                    while PING in d_q:
                        d_q.remove(PING)
                    d_q.insert(0,PING)
                    self.state_map[k] = 16
                    if not self.mode:
                        self._cpipe.send({"state_change":16,"sid":k})
                        with self._mcond:
                            self._mcond.notify()
                self._ping = False
            
            socks = dict(zmq_poll([(connection,zmq.POLLIN | zmq.POLLOUT)], 10))
            ticks += 1
            if time.time() - last > debug_timeout:
                print(socks,ticks)
                last = time.time()

            ##### recv anything ready
            payload_set = list()
            if connection in socks and socks[connection] & zmq.POLLIN:
                last = time.time()
                ticks = 0
                payload_set = connection.recv_all()
            ##### send anything ready
            if connection in socks and socks[connection] & zmq.POLLOUT:
                for sid,d_q in self.reply_map.items():
                    if d_q:
                        msg = d_q.popleft()
                        payload = [sid,b'']
                        if isinstance(msg,list):
                            payload.extend(msg)
                        else:
                            payload.append(msg)
                        connection.send(payload)
            for payload in payload_set:
                if len(payload) > 2:
                    if payload[1] == b'':
                        sid,msg = payload[0],payload[2:]
                        if sid not in self.state_map:
                            if not self.mode:
                                self._cpipe.send({'new_sid':sid})
                                with self._mcond:
                                    self._mcond.notify()
                            self._add_new_sid(sid)
                        if len(msg) == 1 and msg[0]==PING:
                            self.reply(sid,PONG,1)
                            print("____pinged____")
                        elif len(msg) == 1 and msg[0]==PONG:
                            self.state_map[sid] = 1
                            if not self.mode:
                                self._cpipe.send({"state_change":1,"sid":sid})
                                with self._mcond:
                                    self._mcond.notify()
                            print("____ponged____")
                        else:
                            if not self.mode:
                                self._cpipe.send({'task':msg,'sid':sid})
                                with self._mcond:
                                    self._mcond.notify()
                            elif callable(self._task_callback):
                                self._task_callback((sid,msg))
                            else:
                                self.task_map[sid].append(msg)
                    else:
                        print("hmm,separator not found",payload[:2])
                else:
                    print("hmm, message structure is too short:",payload)
        else:
            connection.close()
            self.endpoint = None
            if self._cpipe is not None and not self._cpipe.closed():
                try:
                    self._cpipe.close()
                except:
                    pass

    def _add_new_sid(self,sid):
        if sid in self.state_map:
            return
        self.state_map[sid] = 0
        self.task_map[sid] = deque()
        self.reply_map[sid] = deque()

    def _monitor_pipe(self):
        if os.getpid() != self._pid:
            print("only works in object process")
            return
        if self.mode:
            print("only works in process mode")
            return
        while not self._lthread.check_for_stop():
            with self._mcond:
                if self._mcond.wait(0.1):
                    payload = self._mpipe.recv()
                    # print(payload)
                    try:
                        if "state_change" in payload:
                            self.state_map[payload['sid']] = payload['state_change']
                            continue
                        if 'new_sid' in payload:
                            self._add_new_sid(payload['new_sid'])
                            continue
                        if 'task' in payload:
                            sid = payload['sid']
                            msg = payload['task']
                            if callable(self._task_callback):
                                self._task_callback((sid,msg))
                            else:
                                self.task_map[sid].append(msg)
                    except:
                        import traceback
                        traceback.print_exc()
                        print("failed:",payload)

    def wait(self):
        if self.listener is not None:
            self.listener.wait()
            self.listener = None

    def ping(self):
        if self._mpipe is not None:
            self._mpipe.send({'ping':True})
        else:
            self._ping = True
        time.sleep(0.1)

    def send(self,msg:Union[str,bytes]):
        for ep in list(self.state_map):
            self.reply_map[ep].append(msg)
            if self._mpipe is not None:
                to_pipe = {"send":msg,"sid":ep}
                # print("PIPING:",to_pipe)
                self._mpipe.send(to_pipe)

    def reply(self,sid,msg,priority=0):
        if self.mode:
            self._reply(sid,msg,priority)
        if sid not in self.reply_map:
            return
        else:
            to_pipe = {"reply":msg,"sid":sid,"priority":priority}
            self._mpipe.send(to_pipe)


    def _reply(self,sid,msg,priority=0):
        if sid not in self.reply_map:
            return
        if priority == 1:
            self.reply_map[sid].insert(0,msg)
        else:
            self.reply_map[sid].append(msg)





from typing import Union,List,Dict,Tuple,Any,Type,TypedDict,TypeVar
from os import PathLike
import threading
import multiprocessing as mp
import numpy as np
import zmq
from enum import IntFlag

class ConnectionRole(IntFlag): ...

class BetterThread(threading.Thread): ...
class BetterProcess(mp.Process): ...
class BetterConnection(object): ...

class ClientConnection(BetterConnection): ...
class ClientNetworking(object): ...

class ServerConnection(BetterConnection): ...
class ServerNetworking(object): ...


class ClientRequest(object): ...
def stringify_client_request(request:ClientRequest) -> str: ...
def extract_client_request(request:str) -> ClientRequest: ...




from .system_header import *
import traceback
import signal
import time
import os

class BP_STOP(Exception):
    pass

class BetterProcess(mp.Process):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
        self._external_stop_ = threading.Event()
        self._external_stop_.clear()
        self._ack_stop_ = threading.Event()
        self._ack_stop_.clear()
        self._ack_start_ = threading.Event()
        self._ack_start_.clear()
        self._parent_pipe_,self._child_pipe_ = mp.Pipe()
        self._better_stop_reason_ = None
        self._child_exception_ = None
        self._child_pid_ = None

    def __str__(self):
        state = ((int(self._ack_start_.is_set()) << 0)
                +(int(self._ack_stop_.is_set()) << 1)
                +(int(self._external_stop_.is_set() << 2)))
        if state == 0:
            state = "Pending"
        elif state == 1:
            state = "Running"
        elif state in [2,3]:
            state = "Finished"
        elif state in [4,5]:
            state = 'Stopping'
        elif state in [6,7]:
            state = 'Stopped'
        out = f"<BetterProcess: {state}, name: {self._name}, target: {self._target}>"
        return out
    
    def __repr__(self):
        return str(self)


    def stop(self,reason:str=None):
        if not self._external_stop_.is_set():
            self._external_stop_.set()
            if hasattr(signal,"pidfd_send_signal"):
                signal.pidfd_send_signal(os.pidfd_open(self._child_pid_),signal.SIGUSR1)
            else:
                import subprocess,shlex
                command = f'/usr/bin/bash -c "kill -n 10 {str(self._child_pid_)}"'
                proc = subprocess.Popen(shlex.split(command))
                while proc.poll() is None:
                    time.sleep(0.01)
        if reason is not None:
            if self._better_stop_reason_:
                self._better_stop_reason_ = "\n".join([self._better_stop_reason_,reason])
            else:
                self._better_stop_reason_ = reason

    def check_for_stop(self):
        return self._external_stop_.is_set()

    def _stopped(self):
        self._ack_stop_.set()

    def is_stopped(self):
        return self._ack_stop_.is_set()

    def wait(self,timeout=None):
        wait_at = time.time()
        if timeout is None:
            continue_loop = lambda : True
        else:
            continue_loop = lambda : time.time() - wait_at < timeout
        while (continue_loop()):
            if not self.is_alive():
                self._stopped()
                return True
        return False

    def start(self):
        if self._ack_start_.is_set():
            return
        self._ack_start_.set()
        super().start()
        self._child_pid_ = self._parent_pipe_.recv()
        if not isinstance(self._child_pid_,int):
            self._child_exception_ = self._child_pid_
            self._child_pid_ = None

    def run(self):
        def handler(sig,frame):
            raise BP_STOP
        signal.signal(signal.SIGUSR1,handler)
        try:
            self._child_pipe_.send(os.getpid())
            mp.Process.run(self)
            self._child_pipe_.send(None)
        except BP_STOP:
            self._child_pipe_.send("stop caught")
        except Exception as e:
            tb = traceback.format_exc()
            self._child_pipe_.send((e,tb))
            raise
        finally:
            self._stopped()

    @property
    def exception(self):
        if self._parent_pipe_.poll():
            self._child_exception_ = self._parent_pipe_.recv()
        return self._child_exception_

    def close(self):
        self.stop()
        self.wait()
        self.join()

    @property
    def reason(self):
        if self.is_alive():
            return None
        return self._better_stop_reason_



from .better_thread import BetterThread
from .better_process import BetterProcess
from .better_connection import BetterConnection,ConnectionRole

_ = dict([(x.name,x) for x in ConnectionRole])
if 'OFF_' not in _:
    _['OFF_'] = ConnectionRole(0)
locals().update(_)
del _

from .server import ServerConnection,ServerNetworking
from .client import ClientConnection,ClientNetworking

__version__ = '0.0.1.alpha.5'




from .system_header import *
import struct

import socket

class ConnectionRole(IntFlag):
    OFF_=0
    SEND=1
    RECV=2
    UDP_=4
    TCP_=8
    UNIX=16
    MCST=32
    BIND=64
    REUSE=128

class MessageWrapper(object):
    def __init__(self):
        self.buff = None
        self.ready = False
        self.active = None
        self.rem_segs = None
        self.rem_lens = None
        self._excess = bytes()

    @staticmethod
    def wrap(raw_msg:List[bytes]) -> List[bytes]:
        segments = len(raw_msg)
        seglens = [len(x) for x in raw_msg]
        wrapped = [None]*segments
        index = 0
        while segments:
            pad = struct.pack('>HH',index,seglens[index]) if index else struct.pack('>HHH',segments,index,seglens[index])
            wrapped[index] = pad + raw_msg[index]
            index += 1
            segments -= 1
        return wrapped

    def parse(self,buffer:bytes):
        if self.ready:
            return
        buffer = self._excess+buffer
        self._excess = bytes()
        if self.buff is None:
            self.rem_segs = struct.unpack('>H',buffer[:2])[0]
            buffer = buffer[2:]
            self.buff = [None]*self.rem_segs
            self.active = 0
        while len(buffer) and self.rem_segs:
            if self.rem_lens is None:
                header = struct.unpack('>HH',buffer[:4])
                if header[0] not in [self.active,self.active+1]:
                    print("------------desync---------------")
                self.active = header[0]
                self.rem_lens = header[1]
                buffer = buffer[4:]
                if self.active < len(self.buff):
                    self.buff[self.active] = bytes()
            chunk = buffer[:self.rem_lens]
            self.rem_lens -= len(chunk)
            buffer = buffer[len(chunk):]
            if self.active < len(self.buff):
                self.buff[self.active] += chunk
            if self.rem_lens == 0:
                self.rem_lens = None
                self.rem_segs -= 1
        if self.rem_segs == 0:
            if len(buffer):
                self._excess = buffer
            self.ready = True
            self.rem_segs = None
            self.rem_lens = None

    def get(self):
        if not self.ready:
            return []
        buff = [x for x in self.buff]
        self.buff = None
        self.active = None
        return buff if len(buff) > 1 else buff[0]



def zmq_poll(sockets:List[Tuple[Any,int]],timeout:int=None):
    socks = {}
    for sock,flag in sockets:
        socks[sock.fileno()] = sock
    try:
        result = dict(zmq.zmq_poll(sockets,timeout))
    except Exception as e:
        import traceback
        thread = threading.current_thread()
        try:
            print(thread.getName(),'..stopping')
            thread.stop(traceback.format_exc())
        except:
            pass
        return dict()
    better_result = dict()
    for k,v in result.items():
        if isinstance(k,int):
            try:
                better_result[socks[k]] = v
            except:
                print(k,v,socks,better_result)
                raise
        else:
            better_result[k] = v
    return better_result
zmq_poll.__doc__ = "\n".join(["USE_IN BetterThread",zmq.zmq_poll.__doc__])

class BetterConnection(object):
    def __init__(self,role:ConnectionRole,
                      addr:Union[str,List[str]],
                      port:Union[int,str,List[Union[int,str]]],
                      _ctx:zmq.Context=None):
        self.role = role
        self._ctx = _ctx
        self._ctx_owner = _ctx is None
        self._sock = None
        self._conn = None
        if isinstance(addr,list) or isinstance(port,list):
            if role & ConnectionRole.BIND:
                raise ValueError("Cannot bind to multiple connections.")
            if not (isinstance(addr,list) and isinstance(port,list)):
                raise ValueError("Both addr and port must be lists")
            if len(addr)!=len(port):
                raise ValueError("Both addr and port must be lists of the same length")
            self.net = (addr,port)
        else:
            self.net = ([addr],[port])
        self.sendpoint = None
        self.endpoint = None

    def __str__(self):
        state = ((int(self._conn is not None) << 0)
                +(int(self._sock is None) << 1))
        if state == 0:
            state = "Pending"
        elif state == 1:
            state = "Running"
        elif state in [2,3]:
            state = "Closed"
        out = f"<BetterConnection: {state}, fileno: {self.fileno()}>"
        return out

    def __repr__(self):
        return str(self)


    def connect(self):
        if self.role & ConnectionRole.REUSE:
            return self._connect_socket()
        self._ctx = self._ctx if self._ctx else zmq.Context()
        self.protocol = None
        if self.role & (ConnectionRole.RECV | ConnectionRole.SEND):
            if (self.role & (ConnectionRole.RECV | ConnectionRole.SEND) == (ConnectionRole.RECV | ConnectionRole.SEND)):
                self._sock_type = zmq.ROUTER
            else:
                self._sock_type = zmq.SUB if not self.role & ConnectionRole.SEND else zmq.PUB
        if self.role & ConnectionRole.MCST:
            if self.role & ConnectionRole.UDP_:
                self._sock_type = zmq.DGRAM
                self.protocol = 'udp'
                self.role |= ConnectionRole.BIND # must bind
            else:
                self._sock_type = zmq.STREAM
                self.protocol = 'tcp'
        elif self.role & ConnectionRole.UDP_:
            raise NotImplementedError("If you're here, this isn't functional")
            self.protocol = 'udp'
        elif self.role & ConnectionRole.TCP_:
            self.protocol = 'tcp'
        elif self.role & ConnectionRole.UNIX:
            self.protocol = 'ipc'
        if self.protocol is None:
            self.role |= ConnectionRole.TCP_
            self.protocol = 'tcp'
        # print("socket type:",self._sock_type)
        self._sock = self._ctx.socket(self._sock_type)
        endpoint = f"{self.protocol}://{self.net[0][0]}:{self.net[1][0]}"
        if self.role & ConnectionRole.BIND:
            if self._sock_type == zmq.ROUTER:
                self._sock.identity = endpoint.encode()
            if self.role & ConnectionRole.MCST:
                pass
            self._conn = self._sock.bind(endpoint)
            self.endpoint = self._sock.LAST_ENDPOINT
            # print("Socket bind ",end='')
            if self._sock_type in [zmq.DGRAM, zmq.STREAM]:
                self.sendpoint = self.endpoint
        else:
            if self._sock_type == zmq.ROUTER:
                self.sendpoint = [f"{self.protocol}://{a}:{p}" for a,p in zip(self.net[0],self.net[1])]
            endpoints = [f"{self.protocol}://{a}:{p}" for a,p in zip(self.net[0],self.net[1])]
            self._conn = [None]*len(endpoints)
            for idx,ep in enumerate(endpoints):
                self._conn[idx] = self._sock.connect(ep)
                endpoints[idx] = self._sock.LAST_ENDPOINT
            self.endpoint = endpoints
            # print("Socket connect ",end='')
        # print(self.endpoint)
        if self._sock_type == zmq.SUB:
            self._sock.setsockopt_string(zmq.SUBSCRIBE,'')
        return self

    def _connect_socket(self):
        group = self.net[0][0]
        port = self.net[1][0]
        ttl = 4
        self._sock = socket.socket(
            socket.AF_INET,
            socket.SOCK_DGRAM,
            socket.IPPROTO_UDP
        )
        self._sock.setsockopt(
            socket.IPPROTO_IP,
            socket.IP_MULTICAST_TTL,
            ttl
        )
        self._sock.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_REUSEADDR,
            1
        )
        self._sock.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_REUSEPORT,
            1
        )
        self._sock.setblocking(0)

        self._sock.bind((group,port))

        grp_req = struct.pack("4sl",socket.inet_aton(group),socket.INADDR_ANY)
        self._sock.setsockopt(
            socket.IPPROTO_IP,
            socket.IP_ADD_MEMBERSHIP,
            grp_req
        )
        self.sendpoint = (group,port)
        self._sock_type = zmq.DGRAM

    def get_socket(self):
        return self._sock

    def send(self,msg:Union[ClientRequest,str,List[str]],tag:bytes=None,_raw=True):
        try:
            self.sendto(msg,tag=tag,_raw=_raw)
        except RuntimeError as e:
            if "destination" in str(e).lower():
                raise RuntimeError(str(e) + ", use sendto and specify dest:Tuple")
            raise

    def sendto(self,msg:Union[ClientRequest,str,List[str]],dest:Tuple=None,tag:bytes=None,_raw=True):
        # confirm message is useable
        if not isinstance(msg,(ClientRequest,str,bytes)):
            if not isinstance(msg,(list,tuple)):
                raise ValueError(f"Cannot interpret msg of type: {type(msg)}")
            if not all([isinstance(x,(str,bytes)) for x in msg]):
                raise ValueError(f"Cannot interpret iterable message whose values aren't in (str,bytes)")

        dests = None
        if isinstance(msg,(list,tuple)):
            msg = [x.encode() if isinstance(x,str) else x for x in msg]
            if b'' in msg:
                split_on = msg.index(b'')
                dests = msg[:split_on]
                msg = msg[split_on+1:]
        elif isinstance(msg,ClientRequest):
            payload = msg.get_payload()
            split_on = payload.index(b"")
            dests = payload[:split_on]
            msg = payload[split_on+1:]
        else:
            if isinstance(msg,str):
                msg = msg.encode()
            if isinstance(msg,bytes):
                msg = [msg]


        if self._sock_type in [zmq.PUB]:
            dests = [self.endpoint]

        # print(msg,dest,dests)
        if dests is None:
            def parse_str(d):
                if "://" in d:
                    _,d = d.split("://")
                if ":" not in d:
                    raise ValueError("dest must contain at minimum and 'addr:port' format")
                addr,port = d.split(":")
                return addr,port
            def parse_byt(d):
                if b"://" in d:
                    _,d = d.split(b"://")
                if ":" not in d:
                    raise ValueError("dest must contain at minimum and 'addr:port' format")
                addr,port = d.split(b":")
                return addr,port

            if isinstance(dest,(str,bytes,list,tuple)):
                if isinstance(dest,(str,bytes)):
                    if isinstance(dest,str):
                        addr,port = parse_str(dest)
                    else:
                        addr,port = parse_byt(dest)
                    addr = [addr]
                    port = [port]
                elif isinstance(dest,tuple) and len(dest) == 2 and not (
                    (isinstance(dest[0],str) and ":" in dest) or
                    (isinstance(dest[0],bytes) and b":" in dest) or
                    isinstance(dest[0],list)):
                    addr = [dest[0]]
                    port = [dest[1]]
                elif isinstance(dest,tuple) and len(dest) == 2 and isinstance(dest[0],list):
                    addr = dest[0]
                    port = dest[1]
                elif isinstance(dest,(list,tuple)):
                    addr = []
                    port = []
                    for d in dest:
                        if isinstance(d,str):
                            a,p = parse_str(d)
                        else:
                            a,p = parse_byt(d)
                        addr.append(a)
                        port.append(p)
                else:
                    raise ValueError("Couldn't understand 'dest'")
                dests = [f"{self.protocol}://{a if isinstance(a,str) else a.decode()}:{p if isinstance(p,str) else p.decode()}".encode() for a,p in zip(addr,port)]
            elif self.sendpoint is not None:
                dests = [x for x in self.sendpoint] if isinstance(self.sendpoint,list) else [self.sendpoint]
            # else:
            #     raise RuntimeError("Destination address could not be inferred")
            # maybe still further down in the payload?
        # print(msg,dest,dests)
        if isinstance(msg,str):
            payload = [msg.encode()]
        elif isinstance(msg,bytes):
            payload = [msg]
        else:
            def bytify(obj):
                if hasattr(obj,'encode'):
                    return obj.encode()
                elif hasattr(obj,'dumps'):
                    tmp = (b'',f'd:{obj.__class__.__name__}'.encode(),obj.dumps())
                    if not isinstance(tmp[-1],bytes):
                        tmp = tmp[:-1] + (tmp[-1].encode(),)
                    return tmp
                elif hasattr(obj,'to_bytes'):
                    return (b'',f'b:{obj.__class__.__name__}'.encode(),obj.to_bytes())
                else:
                    raise ValueError(f"Don't know how to get bytes from {type(obj)}")
            multimsg = False
            if isinstance(msg,(tuple,list)):
                payload = [x.encode() if hasattr(x,'encode') else x for x in msg]
                for idx in range(len(payload)):
                    if isinstance(payload[idx],ClientRequest):
                        p = msg.get_payload()
                        split_on = p.index(b"")
                        d = p[:split_on]
                        p = p[split_on+1:]
                        payload[idx] = (d,p)
                        if d:
                            multimsg = True
                    elif not isinstance(payload[idx],bytes):
                        payload[idx] = bytify(payload[idx])
            else:
                payload = bytify(payload)
            if multimsg:
                dests = True
        # print(msg,dest,dests)
        if not dests:
            print(self._sock_type)
            raise RuntimeError("Destinations cannot be derived. ")
        # print(msg,dest,dests)
        if dests == True:
            send_msgs = []
            for dest,msg in payload:
                for x in dest:
                    send_msgs.append([x] + ([b''] if self._sock_type in [zmq.ROUTER] else []) + msg)
        elif dests:
            send_msgs = [[x] + ([b''] if self._sock_type in [zmq.ROUTER] else []) + msg for x in dests]

        if len(send_msgs) > 1:
            self.__send_multi_msg(send_msgs)
            return
        elif send_msgs:
            send_msg = send_msgs[0]
        else:
            return
        # print("send_msg:",send_msg)
        # print(msg,dest,dests)
        if isinstance(self._sock,zmq.Socket):
            if self._sock_type in [zmq.DGRAM,zmq.STREAM]:
                if send_msg and send_msg[0] == self.sendpoint:
                    # print("Send point A",send_msg)
                    self._sock.send_multipart(send_msg)
                else:
                    # print("Send point A",[f'{self.sendpoint.decode()}'.encode()]+send_msg)
                    self._sock.send_multipart([f'{self.sendpoint.decode()}'.encode()]+send_msg)
            else:
                # print(send_msg)
                if len(send_msg) > 1:
                    # print("Send point B",send_msg)
                    self._sock.send_multipart(send_msg)
                else:
                    # print("Send point C",send_msg[0])
                    self._sock.send(send_msg[0])
        else:
            # print("Send point D",len(send_msg[1]))
            self._sock.sendto(send_msg[1],self.sendpoint)

    def __send_multi_msg(self,msg_list):
        if not isinstance(self._sock,zmq.Socket):
            print(msg_list)
            raise ValueError("Got here without being zmq")
        ### assumed zmq for this
        for msg in msg_list:
            if len(msg) > 1:
                print("Send point D",len(msg))
                self._sock.send_multipart(msg)
            else:
                print("Send point F",len(msg))
                self._sock.send(msg[0])

    def recv(self,timeout:int=10,tag:str=None,_raw=True):
        # print('timeout =',timeout)
        socks = zmq_poll([(self._sock,zmq.POLLIN)],timeout=timeout)
        handler = MessageWrapper()
        sender = None
        if self._sock in socks or self._sock.fileno() in socks:
            if isinstance(self._sock,zmq.Socket):
                payload = self._sock.recv_multipart()
                # print("raw recv:",payload)
                if self.role & ConnectionRole.MCST:
                    payload = payload[1:]
                if self._sock.socket_type == zmq.ROUTER:
                    sender = payload[0]
                    payload = payload[1:]
                if not _raw:
                    for chunk in payload:
                        handler.parse(chunk)
                    if not handler.ready:
                        print("....")
                    payload = handler.get()
                    payload = [payload] if not isinstance(payload,list) else payload
                payload = ([sender] if sender is not None else []) + payload
                payload = payload if len(payload) > 1 else payload[0]
            elif isinstance(self._sock,socket.socket):
                payload = self._sock.recv(9000)
            return payload
        return None

    def recv_all(self,timeout:int=10,tag:str=None,count:int=0,_raw=True):
        payloads = []
        if isinstance(self._sock, zmq.Socket):
            payload = self.recv(timeout=timeout,tag=tag,_raw=_raw)
            while payload is not None:
                payloads.append(payload)
                payload = self.recv(timeout=timeout,tag=tag,_raw=_raw)
        elif isinstance(self._sock, socket.socket):
            payload = self.recv(timeout=timeout,tag=tag,_raw=_raw)
            while payload is not None:
                payloads.append(payload)
                payload = self.recv(timeout=0,tag=tag,_raw=_raw)
        return payloads

    def close(self):
        if self._sock is not None:
            self._sock.close()
            self._sock = None
        if self._ctx is not None and self._ctx_owner:
            self._ctx.term()

    def fileno(self):
        return -1 if self._sock is None else self._sock.fileno()

